import logging

from spacy.matcher import Matcher
from spacy.attrs import LOWER

from .helpers import (
    validate_types, hash_types, predict_output_word
)

logger = logging.getLogger('yasmin')
logger.setLevel(logging.INFO)


class WSD:

    def __init__(self, nlp, model, model_types, type_cache):
        self.nlp = nlp
        self.model = model
        self.type_cache = type_cache
        self.model_types = model_types

    def disambiguate(
            self, sent, word, types, model_types=None, context_window=3
    ):
        """
        Disambiguates the occurrences of a word/phrase in a sentence based on
        a set of categories represented as sets of words.

        :param sent: sentence
        :type sent: str
        :param word: word to disambiguate
        :type word: str
        :param types: type names
        :type types: list
        :param model_types: all model types type_name->[word_member, ...]
        :type model_types: dict
        :param context_window: context window size
        :type context_window: int
        :return: disambiguation results
        :rtype: list
        """
        model_types = model_types or self.model_types
        validate_types(selected=types, available=model_types.keys())
        matcher = self._make_phrase_matcher(word)
        doc = self.nlp(sent)
        results = []
        matches = matcher(doc)
        new_type_hash = hash_types(model_types)
        type_matrix = self.type_cache.get(new_type_hash)
        if type_matrix is None:
            type_matrix = self._make_type_matrix(
                model_types=model_types, model=self.model
            )
        filtered = self._filter_types(matrix=type_matrix, names=types)

        for _, start, end in matches:
            context, match = self._get_context(
                doc=doc, start=start, end=end, win=context_window
            )
            _type, prob = self._find_sense(
                context=context, type_matrix=filtered
            )
            results.append({
                'type': _type,
                'prob': prob,
                'offsets': [match.start_char, match.end_char]
            })

        return results

    def _find_sense(self, *, context, type_matrix):
        """
        Finds the most likely sense of a word given a context.

        :param context: context words
        :type context: list
        :param type_matrix: mapping types->one-hot vectors of type vocabulary
        :type type_matrix: dict
        :return: selected type, selection probability
        """
        probs = predict_output_word(model=self.model, context=context)
        type_probs = {}
        for t, type_marks in type_matrix.items():
            ln = type_marks.sum()
            type_probs[t] = (probs * type_marks).sum() / ln
        choice_type, raw_prob = max(type_probs.items(), key=lambda x: x[1])
        choice_prob = raw_prob / sum(type_probs.values())

        return choice_type, choice_prob

    def _get_context(self, *, doc, start, end, win):
        # get the phrase
        match = doc[start:end]

        # get token indices for the context
        context_start = start - win if start - win > 0 else 0
        context_end = end + win if end + win < len(doc) else len(doc)

        # slice the context
        context = [t.orth_.lower() for t in doc[context_start:context_end]
                   if self._validate_context(t, match.orth_)]

        return context, match

    @staticmethod
    def _filter_types(*, matrix, names):
        return {typ: vec for typ, vec in matrix.items() if typ in names}

    @staticmethod
    def _validate_context(context_token, target_word):
        return (
            not context_token.is_punct
            and not context_token.orth_ == target_word
        )

    def _make_phrase_matcher(self, phrase):
        lowered = phrase.lower()
        m = Matcher(self.nlp.vocab)
        name = phrase.replace(' ', '_')
        m.add(name, None, [{LOWER: w.orth_} for w in self.nlp(lowered)])
        return m



from hashlib import md5

import numpy as np
import spacy
from spacy.tokenizer import Tokenizer

from .exceptions import ValidationException, OutOfVocabException


def predict_output_word(*, model, context):
    """
    Report the probability distribution of the center word given the context
    words as input to the trained `model`.

    NOTE: This method is a version of the `predict_output_word` method in
    `gensim.models.word2vec.Word2Vec`.

    :param model: the trained word2vec model
    :type model: gensim.models.word2vec.Word2Vec
    :param context: list of context words
    :type context: list
    """
    if not model.negative:
        raise RuntimeError(
            "We have currently only implemented predict_output_word for the "
            "negative sampling scheme, so you need to have run word2vec with "
            "negative > 0 for this to work."
        )

    if not hasattr(model.wv, 'syn0') or not hasattr(model, 'syn1neg'):
        raise RuntimeError(
            "Parameters required for predicting the output words not found.")

    word_vocabs = [model.wv.vocab[w] for w in context if
                   w in model.wv.vocab]
    if not word_vocabs:
        raise OutOfVocabException(
            "All the input context words are out-of-vocabulary for the "
            "current model.")

    word2_indices = [word.index for word in word_vocabs]

    l1 = np.sum(model.wv.syn0[word2_indices], axis=0)
    if word2_indices and model.cbow_mean:
        l1 /= len(word2_indices)

    # propagate hidden -> output and take softmax to get probabilities
    prob_values = np.exp(np.dot(l1, model.syn1neg.T))
    prob_values /= sum(prob_values)

    return prob_values


def parse_custom_types(types):
    """
    Parses curstom types format as sent through the service.

    :param types: curstom types JSON
    :type types: list
    :return: custom types dictionary
    :rtype: dict
    """
    model_types = {}
    for typ in types:
        name = typ['name']
        model_types[name] = [
            x.lower().strip() for x in typ['keywords'].split(',')
        ]
    return model_types


def validate_types(*, selected, available):
    """
    Makes sure the selected types for a particular phrase/word are in the list
    of declared types.

    :param selected: types selected for disambiguation
    :type selected: iterable
    :param available: all model types
    :type available: iterable
    :raises ValidationException: if one or more of the selected types is not
        in the list of model types
    """
    spurious_types = set(selected) - set(available)
    if spurious_types:
        raise ValidationException(f'Type(s): "{spurious_types}" unknown')


def hash_types(types):
    m = md5()
    types_str = '+'.join(f"{k}:{sorted(set(v))}" for k, v in types.items())
    m.update(types_str.encode('utf-8'))
    return m.digest()


def make_type_matrix(*, model_types, model):
    """
    Produces a dictionary mapping of type names to one-hot vectors of the
    size of the `model` vocabulary, representing the terms present in the
    type according to `model_types`.

    :param model_types: dictionary mapping of type names and term members
    :type model_types: dict
    :param model: gensim word2vec model
    :type model: gensim.models.word2vec.Word2Vec
    :return: type one-hot matrix
    :rtype: dict
    """
    matrix = {}
    vocab = model.wv.vocab
    ln = len(vocab)
    for typ, terms in model_types.items():
        indices = [vocab[term].index for term in terms if term in vocab]
        v = np.zeros(ln)
        np.put(v, indices, np.ones(len(indices)))
        matrix[typ] = v
    return matrix


def custom_tokenizer(nlp):  # pragma: no cover
    """
    Creates a custom spaCy tokenizer.

    :param nlp: a spaCy Language object
    :return:
    """
    prefix_re = spacy.util.compile_prefix_regex(nlp.Defaults.prefixes)
    suffix_re = spacy.util.compile_suffix_regex(nlp.Defaults.suffixes)
    infix_re = spacy.util.compile_infix_regex(nlp.Defaults.infixes)
    tokenizer = Tokenizer(nlp.vocab, nlp.Defaults.tokenizer_exceptions,
                          prefix_re.search, suffix_re.search,
                          infix_re.finditer, token_match=None)

    def make_doc(text): return tokenizer(text)
    return make_doc



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()


setup(
    name='yasmin',
    version='0.1.0',
    packages=find_packages('.', exclude=('tests',)),
    zip_safe=True,
    include_package_data=False,
    description='',
    author='Babylon Health',
    license='Proprietary',
    long_description=(
        'https://github.com/Babylonpartners/yasmin'
    ),
    install_requires=install_requires,
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP'
    ]
)
